""" in-wheel-motored vehicle """


class InWheelMotoredVehicle:
    """ class to define in-wheel-motored vehicle model """
    def __init__(self) -> None:
        # under developing
        raise NotImplementedError



#  Copyright (c) 2019-2020, Mechanical Simulation Corporation
# All rights reserved.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import ctypes
import sys,string,os,re


class vs_solver:
    def __init__(self):
        self.dll_handle = None

    def get_api(self, dll_handle):
        self.dll_handle = dll_handle

        if dll_handle.vs_run is not None and \
                        dll_handle.vs_initialize is not None and \
                        dll_handle.vs_read_configuration is not None and \
                        dll_handle.vs_integrate_io is not None and \
                        dll_handle.vs_copy_export_vars is not None and \
                        dll_handle.vs_terminate_run is not None and \
                        dll_handle.vs_error_occurred is not None and \
                        dll_handle.vs_set_opt_error_dialog is not None and \
                        dll_handle.vs_get_error_message is not None and \
                        dll_handle.vs_road_l is not None:
            dll_handle.vs_road_l.restype = ctypes.c_double
            #vsimports = (ctypes.c_double*1)()
            #vsexports = (ctypes.c_double*1)()
            #dll_handle.vs_integrate_io.argtypes = [ctypes.c_double, ctypes.byref(vsimports), ctypes.byref(vsexports)]
            ##dll_handle.vs.integrate_io.argtypes = [ctypes.c_double, ctypes.c_int]
            return True
        else:
            return False

    def get_char_pointer(self, python_string):
        # python version is greater or equal to 3.0 then we need to define the encoding when converting a string to
        # bytes. Once that is done we can convert the the python string to a char*.
        if sys.version_info >= (3, 0):
            char_pointer = ctypes.c_char_p(bytes(python_string, 'UTF-8'))
        else:
            char_pointer = ctypes.c_char_p(bytes(python_string))
        return char_pointer

    def get_parameter_value(self, line):
        index = line.find(' ')
        if index >= 0:
            return line[index:].strip()
        else:
            return None

    def get_dll_path(self, path_to_sim_file):
        dll_path = None
        prog_dir = None
        veh_code = None
        product_name = None
        product_ver = None
        library_name = None
        bitness_suffix = '_64' if ctypes.sizeof(ctypes.c_voidp) == 8 else '_32'
        platform = sys.platform

        sim_file = open(path_to_sim_file, 'r')

        if (platform == 'linux'):
            libfiletype = 'SOFILE'
        else:
            libfiletype = 'DLLFILE'

        for line in sim_file:
            if line.lstrip().startswith('PROGDIR'):
                prog_dir = self.get_parameter_value(line)
                if prog_dir == '.':
                  prog_dir = os.getcwd()
            elif line.lstrip().startswith(libfiletype):
                dll_path = self.get_parameter_value(line)
            elif line.lstrip().startswith('VEHICLE_CODE'):
                veh_code = self.get_parameter_value(line)
            elif line.lstrip().startswith('PRODUCT_ID'):
                product_name = self.get_parameter_value(line)
            elif line.lstrip().startswith('PRODUCT_VER'):
                product_ver = self.get_parameter_value(line)

        sim_file.close()

        if "tire" in veh_code:
            if platform == 'linux':
              library_name = 'libtire.so.%s'%(product_ver)
            else:
              library_name = "tire" + bitness_suffix
        elif product_name == "CarSim":
            if platform == 'linux':
              library_name = 'libcarsim.so.%s'%(product_ver)
            else:
              library_name = "carsim" + bitness_suffix
        elif product_name == "TruckSim":
            if platform == 'linux':
              library_name = 'libtrucksim.so.%s'%(product_ver)
            else:
              library_name = "trucksim" + bitness_suffix
        elif product_name == "BikeSim":
            if platform == 'linux':
              library_name = 'libbikesim.so.%s'%(product_ver)
            else:
              library_name = "bikesim" + bitness_suffix
        else:
            if platform == 'linux':
              library_name = 'libcarsim.so.%s'%(product_ver)
            else:
              library_name = veh_code + bitness_suffix

        if dll_path is None:
            if sys.platform == 'linux':
              dll_path = os.path.join(prog_dir, library_name); 
            else:
              dll_path = os.path.join(prog_dir, "Programs", "Solvers", library_name + ".dll")
        return dll_path

    def run(self, path_to_sim_file):
        error_occurred = 1
        path_to_sim_file_ptr = self.get_char_pointer(path_to_sim_file)

        if path_to_sim_file_ptr is not None:
            error_occurred = self.dll_handle.vs_run(path_to_sim_file_ptr)

        return error_occurred
        
    def print_error(self):
      self.dll_handle.vs_get_error_message.restype = ctypes.c_char_p
      error_string = ctypes.c_char_p(self.dll_handle.vs_get_error_message())
      print(error_string.value.decode('ascii'))

    def read_configuration(self, path_to_sim_file):
        path_to_sim_file_ptr = self.get_char_pointer(path_to_sim_file)
        platform = sys.platform
        if path_to_sim_file_ptr is not None:
            if platform == 'linux':
              ref_n_import = ctypes.c_long()
              ref_n_export = ctypes.c_longlong()
            else:
              ref_n_import = ctypes.c_int32()
              ref_n_export = ctypes.c_int64()
            ref_t_start = ctypes.c_double()
            ref_t_stop = ctypes.c_double()
            ref_t_step = ctypes.c_double()
            self.dll_handle.vs_read_configuration(path_to_sim_file_ptr,
                                                  ctypes.byref(ref_n_import),
                                                  ctypes.byref(ref_n_export),
                                                  ctypes.byref(ref_t_start),
                                                  ctypes.byref(ref_t_stop),
                                                  ctypes.byref(ref_t_step))
            configuration = {'n_import': ref_n_import.value,
                             'n_export': ref_n_export.value,
                             't_start': ref_t_start.value,
                             't_stop': ref_t_stop.value,
                             't_step': ref_t_step.value}
            return configuration

    def copy_export_vars(self, n_export):
        export_array = (ctypes.c_double * n_export)()
        self.dll_handle.vs_copy_export_vars(ctypes.cast(export_array, ctypes.POINTER(ctypes.c_double)))
        export_list = [export_array[i] for i in range(n_export)]
        return export_list

    def get_road_l(self, x, y):
        x_c_double = ctypes.c_double(x)
        y_c_double = ctypes.c_double(y)
        c_double_return = self.dll_handle.vs_road_l(x_c_double, y_c_double)
        return float(c_double_return)

    def integrate_io(self, t_current, import_array, export_array):
        t_current_c_double = ctypes.c_double(t_current)
        import_c_double_array = (ctypes.c_double * len(import_array))(*import_array)
        export_c_double_array = (ctypes.c_double * len(export_array))(*export_array)

        c_integer_return = self.dll_handle.vs_integrate_io(t_current_c_double,
                                                           ctypes.byref(import_c_double_array),
                                                           ctypes.byref(export_c_double_array))

        export_array = [export_c_double_array[i] for i in range(len(export_array))]
        return c_integer_return, export_array

    def initialize(self, t):
        t_c_double = ctypes.c_double(t)
        self.dll_handle.vs_initialize(t_c_double)

    def terminate_run(self, t):
        t_c_double = ctypes.c_double(t)
        self.dll_handle.vs_terminate_run(t_c_double)



""" normal vehicle model """
import os
from rich import print
from rich.panel import Panel
from rich.text import Text
from pycarsimlib.logger import initialize_logging
logger = initialize_logging(__name__)


class NormalVehicle:
    """ class to define normal vehicle model """
    def __init__(self) -> None:
        logger.info("Vehicle type : NormalVehicle")
        self._define_import_params()
        self._define_export_params()
        self.echo_command_for_carsim_gui()

    def _define_import_params(self):
        """ define control inputs """
        self.import_label_array = [
            "IMP_STEER_SW",
            "IMP_FBK_PDL",
            "IMP_THROTTLE_ENGINE"
        ]

    def _define_export_params(self):
        """ define variables to be observed """
        self.export_label_array = [
            "XCG_TM",
            "YCG_TM",
            "Yaw"
        ]

    def get_import_labels(self):
        """ getter of import value labels """
        return self.import_label_array.copy()

    def get_export_labels(self):
        """ getter of export value labels """
        return self.export_label_array.copy()

    def echo_command_for_carsim_gui(self):
        """ input those commands into 'Additional Data' in the carsim gui. """
        logger.warn("Do not forget to set following variables on carsim gui.")

        # generate command for carsim gui
        _cmd = os.linesep

        # add commands for import variables
        for imp in self.import_label_array:
            _cmd += "IMPORT " + imp + " Replace 0.0! 1" + os.linesep

        # add commands for export variables
        for exp in self.export_label_array:
            _cmd += "EXPORT " + exp + " Replace 0.0! 1" + os.linesep

        panel = Panel(Text(_cmd))
        print(panel)




""" setting up a logger for the multi-agent package """
from logging import basicConfig, DEBUG, INFO, WARNING, ERROR, CRITICAL, Logger, getLogger, FileHandler, Formatter
from rich.logging import RichHandler

"""Note : Print Levels
1. CRITICAL
2. ERROR
3. WARNING
4. INFO
5. DEBUG
"""

def initialize_logging(name: str = __name__) -> Logger:
    """Initialize logger for logging system.

    Args:
        name (str, optional): Name of logging file.
            - Defaults to __name__.

    Returns:
        Logger: Logger to handle.
    """

    # logger settings
    FORMAT = "%(message)s"
    DATE_FORMAT = "%m/%d,%H:%M:%S.%f"
    basicConfig(
        level=INFO,
        format=FORMAT,
        datefmt=DATE_FORMAT,
        handlers=[RichHandler(rich_tracebacks=True)],
    )
    logger = getLogger(name)

    return logger



""" core functions of pycarsimlib """

import os
import sys
import struct
import platform
import shutil
from pycarsimlib.api import vs_solver
from datetime import timedelta
from ctypes import cdll
from typing import Union, List, Dict, Any
from pycarsimlib.logger import initialize_logging
from pycarsimlib.models.normal_vehicle import NormalVehicle
from pycarsimlib.models.in_wheel_motored_vehicle import InWheelMotoredVehicle
logger = initialize_logging(__name__)


class CarsimManager:
    """ carsim manager """
    def __init__(
        self,
        carsim_db_dir: str,
        vehicle_type: str,
        **kwargs
    ) -> None:

        # set simfile path
        self.carsimdb_dir = carsim_db_dir
        self.simfile_path = os.path.join(self.carsimdb_dir, "simfile.sim")
        logger.info(f"simfile path : {self.simfile_path}")

        # select vehicle models
        if vehicle_type == "normal_vehicle":
            self.vehicle = NormalVehicle(**kwargs)
        elif vehicle_type == "in_wheel_motored_vehicle":
            self.vehicle = InWheelMotoredVehicle(**kwargs)
        else:
            logger.error(f"Invalid vehicle type '{vehicle_type}' is specified. ")
            raise AttributeError

        # initialize carsim solver
        self._init_carsim()

        # announce
        logger.info("Carsim solver has been initialized successfully.")

    def _init_carsim(self):
        """ init carsim """
        # get solver instance
        self.solver_api = vs_solver.vs_solver()
        self.path_to_vs_dll = self.solver_api.get_dll_path(self.simfile_path)

        if not self._is_system_available():
            logger.error("Carsim library is invalid.")
            raise RuntimeError

        if not self._is_solver_available():
            logger.error("Carsim solver is invalid.")
            raise RuntimeError

        # now, carsim solver is ready to use. loading params from simfile.
        self.configuration = self.solver_api.read_configuration(self.simfile_path)
        self.t_current = self.configuration.get('t_start')
        self.t_step = self.configuration.get('t_step')

        # prepare importation and exportation of variables
        self.num_of_import_values = self.configuration.get('n_import')
        self.num_of_export_values = self.configuration.get('n_export')
        self.import_label_array = self.vehicle.get_import_labels()
        self.export_label_array = self.vehicle.get_export_labels()
        self.export_array = self.solver_api.copy_export_vars(self.num_of_export_values)

    def _is_system_available(self):
        """ check if the carsim library is available in the system. """
        # dealing with different os
        current_os = platform.system()
        if self.path_to_vs_dll is not None \
                and os.path.exists(self.path_to_vs_dll):

            if current_os == "Linux":
                mc_type = platform.machine()
                if mc_type == 'x86_64':
                    dll_size = 64
                else:
                    dll_size = 32
            else:
                if "_64" in self.path_to_vs_dll:
                    dll_size = 64
                else:
                    dll_size = 32

        # check availability
        _system_word_size = (8 * struct.calcsize("P"))  # 32 or 64
        if _system_word_size != dll_size:
            print("Python size (32/64) must match size of .dlls being loaded.")
            print("Python size:", _system_word_size, "DLL size:", dll_size)
            return False
        else:
            return True

    def _is_solver_available(self):
        """ is solver available """
        try:
            self.vs_dll = cdll.LoadLibrary(self.path_to_vs_dll)
            if self.vs_dll is not None:
                if self.solver_api.get_api(self.vs_dll):
                    return True
        except Exception as err_msg:
            logger.error(err_msg)
            return False

    def reset(self):
        """ reset """
        # under developing
        raise NotImplementedError

    def step(self, action, delta_time):
        """ step run of carsim for delta_time with given action """
        _loop_num = delta_time // timedelta(seconds=self.t_step)

        for _ in range(_loop_num):

            # Run the integration loop
            self.t_current += self.t_step  # increment the time

            # reshape dict to list
            import_array = []
            for key in self.import_label_array:
                if key in action:
                    import_array.append(action[key])
                else:
                    logger.error("Invalid format of carsim action.")
                    raise AttributeError

            # call vs_api function
            status, self.export_array = self.solver_api.integrate_io(self.t_current, import_array, self.export_array)

        return dict(zip(self.export_label_array, self.export_array)), status, self.t_current

    def run_all(self):
        """ run all simulation steps at once """
        error_occurred = 1
        logger.info("##### Run all simulation steps #####")
        error_occurred = self.solver_api.run(
            self.simfile_path.replace('\\\\', '\\')
        )
        if error_occurred != 0:
            print("ERROR OCCURRED:  ")
            self.solver_api.print_error()
            sys.exit(error_occurred)
        logger.info("##### End of simulation #####")

    def close(self):
        """ close carsim solver normally"""
        self.solver_api.terminate_run(self.t_current)
        logger.info("Carsim solver terminated normally.")

    def save_results_into_carsimdb(self, results_source_dir="", results_target_dir=""):
        """ copy latest results dir into carsimdb location """

        # keep arguments
        _source_dir = results_source_dir
        _target_dir = results_target_dir

        # set paths
        if not _source_dir:
            _source_dir = os.path.join(os.getcwd(), "Results")
        if not _target_dir:
            _target_dir = os.path.join(self.carsimdb_dir, "Results")

        # copy results dir
        logger.info(f"Saving results into {_target_dir}")
        shutil.copytree(_source_dir, _target_dir, dirs_exist_ok=True)
        logger.info("Successfully saved results.")



""" main process of pycarsimlib"""
import pycarsimlib
from pycarsimlib.logger import initialize_logging
logger = initialize_logging(__name__)


def main() -> None:
    """ echo version """
    logger.info(f"pycarsimlib version=={pycarsimlib.__version__}")



""" setup script of pycarsimlib """
import setuptools
import pycarsimlib

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="pycarsimlib",
    version=pycarsimlib.__version__,
    install_requires=[
        "typing>=3.7.4.3",
        "rich>=12.6.0",
        "numpy>=1.19.5",
    ],
    entry_points={
        'console_scripts': [
        ],
    },
    author="MizuhoAOKI",
    author_email="mizuhoaoki1998@gmail.com",
    description="pycarsimlib: carsim wrapper",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/MizuhoAOKI/pycarsimlib",
    download_url="https://github.com/MizuhoAOKI/pycarsimlib",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Topic :: Scientific/Engineering",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.8',
)
